@@ -53,14 +53,30 @@ import static com.openexchange.java.Autoboxing.I;
 import java.net.URI;
 import java.net.URISyntaxException;
 import java.util.Collection;
-import java.util.HashMap;
+import java.util.Iterator;
+import java.util.LinkedHashSet;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
+import java.util.Set;
+import java.util.Map.Entry;
+import java.util.concurrent.CancellationException;
+import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.ExecutionException;
+import java.util.concurrent.Future;
 import org.slf4j.Logger;
+import com.hazelcast.core.Cluster;
+import com.hazelcast.core.HazelcastInstance;
+import com.hazelcast.core.IExecutorService;
+import com.hazelcast.core.Member;
 import com.openexchange.config.ConfigurationService;
+import com.openexchange.dovecot.doveadm.client.DoveAdmClient;
 import com.openexchange.exception.OXException;
+import com.openexchange.java.Streams;
 import com.openexchange.java.Strings;
+import com.openexchange.lock.AccessControl;
+import com.openexchange.lock.LockService;
+import com.openexchange.lock.ReentrantLockAccessControl;
 import com.openexchange.push.PushExceptionCodes;
 import com.openexchange.push.PushListener;
 import com.openexchange.push.PushListenerService;
@@ -70,11 +86,18 @@ import com.openexchange.push.PushUserInfo;
 import com.openexchange.push.PushUtility;
 import com.openexchange.push.dovecot.locking.DovecotPushClusterLock;
 import com.openexchange.push.dovecot.locking.SessionInfo;
+import com.openexchange.push.dovecot.registration.RegistrationContext;
+import com.openexchange.push.dovecot.registration.RegistrationContext.DoveAdmClientProvider;
 import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
+import com.openexchange.session.ObfuscatorService;
 import com.openexchange.session.Session;
 import com.openexchange.sessiond.SessiondService;
-import com.openexchange.sessiond.SessiondServiceExtended;
+import com.openexchange.sessionstorage.hazelcast.serialization.PortableMultipleSessionRemoteLookUp;
+import com.openexchange.sessionstorage.hazelcast.serialization.PortableSession;
+import com.openexchange.sessionstorage.hazelcast.serialization.PortableSessionCollection;
+import com.openexchange.threadpool.AbstractTask;
+import com.openexchange.threadpool.ThreadPoolService;
 
 
 /**
@@ -137,6 +160,7 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
     private final String authLogin;
     private final String authPassword;
     private final URI uri;
+    private final AccessControl globalLock;
 
     /**
      * Initializes a new {@link DovecotPushManagerService}.
@@ -146,7 +170,7 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
         name = "Dovecot Push Manager";
         this.services = services;
         this.clusterLock = clusterLock;
-        listeners = new HashMap<SimpleKey, DovecotPushListener>(512, 0.9F);
+        listeners = new ConcurrentHashMap<SimpleKey, DovecotPushListener>(512, 0.9F, 1);
 
         // Parse auth data
         {
@@ -171,6 +195,28 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                 throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, null == endPoint ? "<empty>" : endPoint);
             }
         }
+
+        // The fall-back lock
+        globalLock = new ReentrantLockAccessControl();
+    }
+
+    @Override
+    public String toString() {
+        return name;
+    }
+
+    private AccessControl getlockFor(int userId, int contextId) {
+        LockService lockService = services.getOptionalService(LockService.class);
+        if (null == lockService) {
+            return globalLock;
+        }
+
+        try {
+            return lockService.getAccessControlFor("dovecotpush", 1, userId, contextId);
+        } catch (Exception e) {
+            LOGGER.warn("Failed to acquire lock for user {} in context {}. Using global lock instead.", I(userId), I(contextId), e);
+            return globalLock;
+        }
     }
 
     @Override
@@ -192,40 +238,174 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
         // Prefer permanent listener prior to performing look-up for another valid session
         if (hasPermanentPush(userId, contextId)) {
             try {
-                Session session = generateSessionFor(userId, contextId);
-                return injectAnotherListenerUsing(session, true);
+                DoveAdmClient doveAdmClient = services.getOptionalService(DoveAdmClient.class);
+                RegistrationContext registrationContext;
+                if (null == doveAdmClient) {
+                    Session session = generateSessionFor(userId, contextId);
+                    registrationContext = RegistrationContext.createSessionContext(session);
+                } else {
+                    registrationContext = RegistrationContext.createDoveAdmClientContext(userId, contextId, new ServiceLookupDoveAdmClientProvider(services));
+                }
+
+                return injectAnotherListenerUsing(registrationContext, true);
             } catch (OXException e) {
                 // Failed to inject a permanent listener
             }
         }
 
+        Session session = lookUpSessionFor(userId, contextId, oldSession);
+        if (null == session) {
+            // No push-capable session
+            return null;
+        }
+
+        DoveAdmClient doveAdmClient = services.getOptionalService(DoveAdmClient.class);
+        RegistrationContext registrationContext = null == doveAdmClient ? RegistrationContext.createSessionContext(session) : RegistrationContext.createDoveAdmClientContext(userId, contextId, new ServiceLookupDoveAdmClientProvider(services));
+        return injectAnotherListenerUsing(registrationContext, false);
+    }
+
+    /**
+     * Looks-up a push-capable session for specified user.
+     *
+     * @param userId The user identifier
+     * @param contextId The context identifier
+     * @param optOldSession The optional old session (which expired)
+     * @return A push-capable session or <code>null</code>
+     */
+    public Session lookUpSessionFor(int userId, int contextId, Session optOldSession) {
         // Look-up sessions
         SessiondService sessiondService = services.getService(SessiondService.class);
         if (null != sessiondService) {
-            String oldSessionId = oldSession.getSessionID();
+            String oldSessionId = null == optOldSession ? null : optOldSession.getSessionID();
 
-            // Query local ones first
+            // Query local ones
             Collection<Session> sessions = sessiondService.getSessions(userId, contextId);
             for (Session session : sessions) {
-                if (!oldSessionId.equals(session.getSessionID()) && PushUtility.allowedClient(session.getClient(), session, true)) {
-                    return injectAnotherListenerUsing(session, false);
-                }
-            }
-
-            // Look-up remote sessions, too, if possible
-            if (sessiondService instanceof SessiondServiceExtended) {
-                sessions = ((SessiondServiceExtended) sessiondService).getSessions(userId, contextId, true);
-                for (Session session : sessions) {
-                    if (!oldSessionId.equals(session.getSessionID()) && PushUtility.allowedClient(session.getClient(), session, true)) {
-                        return injectAnotherListenerUsing(session, false);
-                    }
+                if ((null == oldSessionId || false == oldSessionId.equals(session.getSessionID())) && PushUtility.allowedClient(session.getClient(), session, true)) {
+                    return session;
                 }
             }
         }
 
+        // Look-up remote sessions, too, if possible
+        return lookUpRemoteSessionFor(userId, contextId, optOldSession);
+    }
+
+    private Session lookUpRemoteSessionFor(int userId, int contextId, Session optOldSession) {
+        HazelcastInstance hzInstance = services.getOptionalService(HazelcastInstance.class);
+        ObfuscatorService obfuscatorService = services.getOptionalService(ObfuscatorService.class);
+        if (null == hzInstance || null == obfuscatorService) {
+            return null;
+        }
+
+        Cluster cluster = hzInstance.getCluster();
+
+        // Get local member
+        Member localMember = cluster.getLocalMember();
+
+        // Determine other cluster members
+        Set<Member> otherMembers = getOtherMembers(cluster.getMembers(), localMember);
+        if (otherMembers.isEmpty()) {
+            return null;
+        }
+
+        IExecutorService executor = hzInstance.getExecutorService("default");
+        Map<Member, Future<PortableSessionCollection>> futureMap = executor.submitToMembers(new PortableMultipleSessionRemoteLookUp(userId, contextId), otherMembers);
+        String oldSessionId = null == optOldSession ? null : optOldSession.getSessionID();
+        for (Iterator<Entry<Member, Future<PortableSessionCollection>>> it = futureMap.entrySet().iterator(); it.hasNext();) {
+            Future<PortableSessionCollection> future = it.next().getValue();
+            // Check Future's return value
+            int retryCount = 3;
+            while (retryCount-- > 0) {
+                try {
+                    PortableSessionCollection portableSessionCollection = future.get();
+                    retryCount = 0;
+
+                    PortableSession[] portableSessions = portableSessionCollection.getSessions();
+                    if (null != portableSessions) {
+                        for (PortableSession portableSession : portableSessions) {
+                            portableSession.setPassword(obfuscatorService.unobfuscate(portableSession.getPassword()));
+                            if ((null == oldSessionId || false == oldSessionId.equals(portableSession.getSessionID())) && PushUtility.allowedClient(portableSession.getClient(), portableSession, true)) {
+                                cancelRest(it);
+                                return portableSession;
+                            }
+                        }
+                    }
+                } catch (InterruptedException e) {
+                    // Interrupted - Keep interrupted state
+                    Thread.currentThread().interrupt();
+                } catch (CancellationException e) {
+                    // Canceled
+                    retryCount = 0;
+                } catch (ExecutionException e) {
+                    Throwable cause = e.getCause();
+
+                    // Check for Hazelcast timeout
+                    if (!(cause instanceof com.hazelcast.core.OperationTimeoutException)) {
+                        if (cause instanceof RuntimeException) {
+                            throw ((RuntimeException) cause);
+                        }
+                        if (cause instanceof Error) {
+                            throw (Error) cause;
+                        }
+                        throw new IllegalStateException("Not unchecked", cause);
+                    }
+
+                    // Timeout while awaiting remote result
+                    if (retryCount <= 0) {
+                        // No further retry
+                        cancelFutureSafe(future);
+                    }
+                }
+            }
+        }
         return null;
     }
 
+    /**
+     * Gets the other cluster members
+     *
+     * @param allMembers All known members
+     * @param localMember The local member
+     * @return Other cluster members
+     */
+    private Set<Member> getOtherMembers(Set<Member> allMembers, Member localMember) {
+        Set<Member> otherMembers = new LinkedHashSet<Member>(allMembers);
+        if (!otherMembers.remove(localMember)) {
+            LOGGER.warn("Couldn't remove local member from cluster members.");
+        }
+        return otherMembers;
+    }
+
+    /**
+     * Cancels given {@link Future} safely
+     *
+     * @param future The {@code Future} to cancel
+     */
+    static <V> void cancelFutureSafe(Future<V> future) {
+        if (null != future) {
+            try { future.cancel(true); } catch (Exception e) {/*Ignore*/}
+        }
+    }
+
+    private <V> void cancelRest(final Iterator<Entry<Member, Future<PortableSessionCollection>>> it) {
+        ThreadPoolService threadPool = services.getOptionalService(ThreadPoolService.class);
+        if (null != threadPool) {
+            AbstractTask<Void> task = new AbstractTask<Void>() {
+
+                @Override
+                public Void call() throws Exception {
+                    while (it.hasNext()) {
+                        Future<PortableSessionCollection> future = it.next().getValue();
+                        cancelFutureSafe(future);
+                    }
+                    return null;
+                }
+            };
+            threadPool.submit(task);
+        }
+    }
+
     /**
      * Tries to look-up another valid session and injects a new listener for it (discarding the existing one bound to given <code>oldSession</code>)
      *
@@ -234,27 +414,23 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
      * @return The new listener or <code>null</code>
      * @throws OXException If operation fails
      */
-    public DovecotPushListener injectAnotherListenerUsing(Session newSession, boolean permanent) {
-        synchronized (this) {
-            DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, newSession, permanent, this, services);
-            // Replace old/existing one
-            listeners.put(SimpleKey.valueOf(newSession), listener);
-            return listener;
-        }
+    public DovecotPushListener injectAnotherListenerUsing(RegistrationContext registrationContext, boolean permanent) {
+        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, registrationContext, permanent, this, services);
+        // Replace old/existing one
+        listeners.put(SimpleKey.valueOf(registrationContext.getUserId(), registrationContext.getContextId()), listener);
+        return listener;
     }
 
     private void stopAll() {
-        synchronized (this) {
-            for (Map.Entry<SimpleKey, DovecotPushListener> entry : listeners.entrySet()) {
-                try {
-                    entry.getValue().unregister(false);
-                } catch (Exception e) {
-                    SimpleKey key = entry.getKey();
-                    LOGGER.warn("Failed to stop listener for user {} in context {}", key.userId, key.contextId);
-                }
+        for (Map.Entry<SimpleKey, DovecotPushListener> entry : listeners.entrySet()) {
+            try {
+                entry.getValue().unregister(false);
+            } catch (Exception e) {
+                SimpleKey key = entry.getKey();
+                LOGGER.warn("Failed to stop listener for user {} in context {}", key.userId, key.contextId);
             }
-            listeners.clear();
         }
+        listeners.clear();
     }
 
     /**
@@ -268,7 +444,10 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
      * @throws OXException If unregistration attempt fails
      */
     public StopResult stopListener(boolean tryToReconnect, boolean stopIfPermanent, int userId, int contextId) throws OXException {
-        synchronized (this) {
+        AccessControl lock = getlockFor(userId, contextId);
+        Runnable cleanUpTask = null;
+        try {
+            lock.acquireGrant();
             SimpleKey key = SimpleKey.valueOf(userId, contextId);
             DovecotPushListener listener = listeners.get(key);
             if (null != listener) {
@@ -280,8 +459,8 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                 listeners.remove(key);
 
                 boolean tryRecon = tryToReconnect || (!listener.isPermanent() && hasPermanentPush(userId, contextId));
-                boolean reconnected = listener.unregister(tryRecon);
-                if (!reconnected) {
+                cleanUpTask = listener.unregister(tryRecon);
+                if (null != cleanUpTask) {
                     return StopResult.STOPPED;
                 }
 
@@ -289,6 +468,14 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                 return (null != newListener && newListener.isPermanent()) ? StopResult.RECONNECTED_AS_PERMANENT : StopResult.RECONNECTED;
             }
             return StopResult.NONE;
+        } catch (InterruptedException e) {
+            Thread.currentThread().interrupt();
+            throw new IllegalStateException(e);
+        } finally {
+            Streams.close(lock);
+            if (null != cleanUpTask) {
+                cleanUpTask.run();
+            }
         }
     }
 
@@ -312,6 +499,28 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
         clusterLock.refreshLock(sessionInfo);
     }
 
+    static class ServiceLookupDoveAdmClientProvider implements DoveAdmClientProvider {
+
+        private final ServiceLookup services;
+
+        /**
+         * Initializes a new {@link DovecotPushManagerService.ServiceLookupDoveAdmClientProvider}.
+         */
+        ServiceLookupDoveAdmClientProvider(ServiceLookup services) {
+            super();
+            this.services = services;
+        }
+
+        @Override
+        public DoveAdmClient getDoveAdmClient() throws OXException {
+            DoveAdmClient doveAdmClient = services.getOptionalService(DoveAdmClient.class);
+            if (null == doveAdmClient) {
+                throw ServiceExceptionCode.absentService(DoveAdmClient.class);
+            }
+            return doveAdmClient;
+        }
+    }
+
     // --------------------------------------------------------------------------------------------------------------------------
 
     @Override
@@ -321,16 +530,21 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
         }
         int contextId = session.getContextId();
         int userId = session.getUserId();
-        SessionInfo sessionInfo = new SessionInfo(session, false);
+        DoveAdmClient doveAdmClient = services.getOptionalService(DoveAdmClient.class);
+        RegistrationContext registrationContext = null == doveAdmClient ? RegistrationContext.createSessionContext(session) : RegistrationContext.createDoveAdmClientContext(userId, contextId, new ServiceLookupDoveAdmClientProvider(services));
+
+        SessionInfo sessionInfo = new SessionInfo(registrationContext, false);
         if (clusterLock.acquireLock(sessionInfo)) {
-            synchronized (this) {
-                // Locked...
-                boolean unlock = true;
+            // Locked...
+            boolean unlock = true;
+            try {
                 boolean removeListener = false;
                 SimpleKey key = SimpleKey.valueOf(userId, contextId);
+                AccessControl lock = getlockFor(userId, contextId);
                 try {
+                    lock.acquireGrant();
                     if (false == listeners.containsKey(key)) {
-                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, session, false, this, services);
+                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, registrationContext, false, this, services);
                         listeners.put(key, listener);
                         removeListener = true;
                         String reason = listener.initateRegistration();
@@ -347,13 +561,18 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                         // Already running for session user
                         LOGGER.info("Did not start Dovecot listener for user {} in context {} with session {} ({}) as there is already an associated listener", I(userId), I(contextId), session.getSessionID(), session.getClient());
                     }
+                } catch (InterruptedException e) {
+                    Thread.currentThread().interrupt();
+                    throw new OXException(e);
                 } finally {
                     if (removeListener) {
                         listeners.remove(key);
                     }
-                    if (unlock) {
-                        releaseLock(sessionInfo);
-                    }
+                    Streams.close(lock);
+                }
+            } finally {
+                if (unlock) {
+                    releaseLock(sessionInfo);
                 }
             }
         } else {
@@ -401,8 +620,7 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
     }
 
     private Session generateSessionFor(int userId, int contextId) throws OXException {
-        PushListenerService pushListenerService = services.getService(PushListenerService.class);
-        return pushListenerService.generateSessionFor(new PushUser(userId, contextId));
+        return generateSessionFor(new PushUser(userId, contextId));
     }
 
     private Session generateSessionFor(PushUser pushUser) throws OXException {
@@ -416,21 +634,30 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
             return null;
         }
 
-        Session session = generateSessionFor(pushUser);
-        int contextId = session.getContextId();
-        int userId = session.getUserId();
+        int contextId = pushUser.getContextId();
+        int userId = pushUser.getUserId();
+        DoveAdmClient doveAdmClient = services.getOptionalService(DoveAdmClient.class);
+        RegistrationContext registrationContext;
+        if (null == doveAdmClient) {
+            Session session = generateSessionFor(pushUser);
+            registrationContext = RegistrationContext.createSessionContext(session);
+        } else {
+            registrationContext = RegistrationContext.createDoveAdmClientContext(userId, contextId, new ServiceLookupDoveAdmClientProvider(services));
+        }
 
-        SessionInfo sessionInfo = new SessionInfo(session, true);
+        SessionInfo sessionInfo = new SessionInfo(registrationContext, true);
         if (clusterLock.acquireLock(sessionInfo)) {
-            synchronized (this) {
-                // Locked...
-                boolean unlock = true;
+            // Locked...
+            boolean unlock = true;
+            try {
                 boolean removeListener = false;
                 SimpleKey key = SimpleKey.valueOf(userId, contextId);
+                AccessControl lock = getlockFor(userId, contextId);
                 try {
+                    lock.acquireGrant();
                     DovecotPushListener current = listeners.get(key);
                     if (null == current) {
-                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, session, true, this, services);
+                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, registrationContext, true, this, services);
                         listeners.put(key, listener);
                         removeListener = true;
                         String reason = listener.initateRegistration();
@@ -442,11 +669,11 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                         }
 
                         // Registration failed
-                        LOGGER.info("Could not register permanent Dovecot listener for user {} in context {} with session {} ({})", I(userId), I(contextId), session.getSessionID(), session.getClient());
+                        LOGGER.info("Could not register permanent Dovecot listener for user {} in context {}", I(userId), I(contextId));
                     } else if (!current.isPermanent()) {
                         // Cancel current & replace
                         current.unregister(false);
-                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, session, true, this, services);
+                        DovecotPushListener listener = new DovecotPushListener(uri, authLogin, authPassword, registrationContext, true, this, services);
                         listeners.put(key, listener);
                         removeListener = true;
                         String reason = listener.initateRegistration();
@@ -458,22 +685,27 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
                         }
 
                         // Registration failed
-                        LOGGER.info("Could not register permanent Dovecot listener for user {} in context {} with session {} ({})", I(userId), I(contextId), session.getSessionID(), session.getClient());
+                        LOGGER.info("Could not register permanent Dovecot listener for user {} in context {}", I(userId), I(contextId));
                     } else {
                         // Already running for session user
-                        LOGGER.info("Did not start permanent Dovecot listener for user {} in context {} with session {} ({}) as there is already an associated listener", I(userId), I(contextId), session.getSessionID(), session.getClient());
+                        LOGGER.info("Did not start permanent Dovecot listener for user {} in context {} as there is already an associated listener", I(userId), I(contextId));
                     }
+                } catch (InterruptedException e) {
+                    Thread.currentThread().interrupt();
+                    throw new OXException(e);
                 } finally {
                     if (removeListener) {
                         listeners.remove(key);
                     }
-                    if (unlock) {
-                        releaseLock(sessionInfo);
-                    }
+                    Streams.close(lock);
+                }
+            } finally {
+                if (unlock) {
+                    releaseLock(sessionInfo);
                 }
             }
         } else {
-            LOGGER.info("Could not acquire lock to start permanent Dovecot listener for user {} in context {} with session {} ({}) as there is already an associated listener", I(userId), I(contextId), session.getSessionID(), session.getClient());
+            LOGGER.info("Could not acquire lock to start permanent Dovecot listener for user {} in context {} as there is already an associated listener", I(userId), I(contextId));
         }
 
         // No listener registered for given session
@@ -504,11 +736,9 @@ public class DovecotPushManagerService implements PushManagerExtendedService {
     @Override
     public List<PushUserInfo> getAvailablePushUsers() throws OXException {
         List<PushUserInfo> l = new LinkedList<PushUserInfo>();
-        synchronized (this) {
-            for (Map.Entry<SimpleKey, DovecotPushListener> entry : listeners.entrySet()) {
-                SimpleKey key = entry.getKey();
-                l.add(new PushUserInfo(new PushUser(key.userId, key.contextId), entry.getValue().isPermanent()));
-            }
+        for (Map.Entry<SimpleKey, DovecotPushListener> entry : listeners.entrySet()) {
+            SimpleKey key = entry.getKey();
+            l.add(new PushUserInfo(new PushUser(key.userId, key.contextId), entry.getValue().isPermanent()));
         }
         return l;
     }
