@@ -49,10 +49,12 @@
 
 package com.openexchange.push.dovecot.commands;
 
+import static com.openexchange.imap.util.ImapUtility.prepareImapCommandForLogging;
 import org.slf4j.Logger;
 import com.openexchange.imap.IMAPCommandsCollection;
 import com.openexchange.imap.IMAPException;
 import com.openexchange.imap.util.ImapUtility;
+import com.openexchange.log.LogProperties;
 import com.sun.mail.iap.BadCommandException;
 import com.sun.mail.iap.CommandFailedException;
 import com.sun.mail.iap.ProtocolException;
@@ -95,12 +97,15 @@ public class UnregistrationCommand implements ProtocolCommand {
             LOGGER.info("Unregistered push notification for {} using: {}", imapFolder.getStore(), command);
             return Boolean.TRUE;
         } else if (response.isBAD()) {
+            LogProperties.putProperty(LogProperties.Name.MAIL_COMMAND, prepareImapCommandForLogging(command));
             throw new BadCommandException(IMAPException.getFormattedMessage(IMAPException.Code.PROTOCOL_ERROR, command, ImapUtility.appendCommandInfo(response.toString(), imapFolder)));
         } else if (response.isNO()) {
+            LogProperties.putProperty(LogProperties.Name.MAIL_COMMAND, prepareImapCommandForLogging(command));
             throw new CommandFailedException(IMAPException.getFormattedMessage(IMAPException.Code.PROTOCOL_ERROR, command, ImapUtility.appendCommandInfo(response.toString(), imapFolder)));
         } else {
+            LogProperties.putProperty(LogProperties.Name.MAIL_COMMAND, prepareImapCommandForLogging(command));
             protocol.handleResult(response);
         }
         return Boolean.FALSE;
     }
-}
\ No newline at end of file
+}
