@@ -135,7 +135,7 @@ public class HzDovecotPushClusterLock extends AbstractDovecotPushClusterLock {
             return true;
         }
 
-        long now = System.nanoTime();
+        long now = System.currentTimeMillis();
         String previous = map.putIfAbsent(key, generateValue(now, sessionInfo));
 
         if (null == previous) {
@@ -144,7 +144,7 @@ public class HzDovecotPushClusterLock extends AbstractDovecotPushClusterLock {
         }
 
         // Check if valid
-        if (validValue(previous, now, hzInstance)) {
+        if (validValue(previous, now)) {
             // Locked
             return false;
         }
@@ -169,7 +169,7 @@ public class HzDovecotPushClusterLock extends AbstractDovecotPushClusterLock {
             return;
         }
 
-        map.put(generateKey(sessionInfo), generateValue(System.nanoTime(), sessionInfo));
+        map.put(generateKey(sessionInfo), generateValue(System.currentTimeMillis(), sessionInfo));
     }
 
     @Override
